"""
See https://www.researchgate.net/publication/271198281_Benchmark_instances_for_the_Multidimensional_Knapsack_Problem 
"""

from pycsp3 import *

profits, wmatrix, capacities, pmatrix = data  # , pmatrix = data
nItems, nBins = len(profits), len(capacities)

# x[i] is 1 if the item i is packed
x = VarArray(size=nItems, dom={0, 1})

# w[j] si the total weight in the jth bin
w = VarArray(size=nBins, dom=lambda j: range(capacities[j] + 1))

z = Var(range(sum(profits) + 1))

satisfy(
    [Knapsack(x, weights=weights, wlimit=w[j], profits=pmatrix[j]) >= z for j, weights in enumerate(wmatrix)],

    # computing the objective value
    z == profits * x
)

maximize(
    # maximizing the profit of packed items
    z
)

"""
1) one can write wcondition=eq(w[j]) or wcondition=le(w[j]) or wlimit=w[j]

2) for being compatible with the competition mini-track, we use:
  [x * weights <= w[j] for j, weights in enumerate(wmatrix)],
  [x * pmatrix[j] >= z for j, weights in enumerate(wmatrix)],

"""
