from pycsp3 import *

weights, k = data
n = len(weights)

# x[i] is the successor node of node i (in the cycle where i belongs)
x = VarArray(size=n, dom=range(n))

# y[i] is the cycle (index) where the node i belongs
y = VarArray(size=n, dom=range(n))

satisfy(
    AllDifferent(x),

    # ensuring correct cycles
    [y[i] == y[x[i]] for i in range(n)],

    # disabling infeasible arcs
    [x[i] != j for i in range(n) for j in range(n) if i != j and weights[i][j] < 0],

    # each cycle has k as maximum length
    BinPacking(y, sizes=1) <= k,

    # tag(symmetry-breaking)
    Precedence(y)
)

maximize(
    # maximizing the sum of arc weights of selected cycles
    Sum(weights[i][x[i]] for i in range(n))
)
