"""
See https://fr.wikipedia.org/wiki/K-m%C3%A9diane
    https://en.wikipedia.org/wiki/K-medians_clustering
    results in Chapter 11 of Constraint-based local search by P. Van Hentenryck and L. Michel
    
Examples of Execution:
  python KMedian.py -data=pmed1.json
  python KMedian.py -variant=aux -data=pmed1.json
  python KMedian.py -data=pmed1.txt -dataparser=KMedian_Parser.py
"""

from pycsp3 import *

distances, k = data
n = len(distances)

# x[i] is the ith selected node
x = VarArray(size=k, dom=range(n))

satisfy(
    # selected nodes must be all different
    AllDifferent(x),

    # tag(symmetry-breaking)
    Increasing(x, strict=True)
)

if not variant():
    minimize(
        # minimizing the minimal distances between nodes and the selected ones
        Sum(Minimum(distances[c][j] for c in x) for j in range(n))
    )

elif variant("aux"):

    # d[i][j] is the distance between the ith selected node and the jth node
    d = VarArray(size=[k, n], dom=distances)

    satisfy(
        # computing distances
        d[i][j] == distances[:, j][x[i]] for i in range(k) for j in range(n)
    )

    minimize(
        # minimizing the minimal distances between nodes and the selected ones
        Sum(Minimum(d[:, j]) for j in range(n))
    )

""" Comments
1) generating the first variant may be very time expensive

2) the two-dimensional array called distances used as domain is automatically transformed into a set of integers
"""
