"""
Pigment Sequencing Problem

See Xavier Gillard, Pierre Schaus:
Large Neighborhood Search with Decision Diagrams. IJCAI 2022: 4754-4760
"""

from pycsp3 import *

nOrders, changeCosts, stockingCosts, demands = data
nItems, horizon = len(demands), len(demands[0])

required = [[sum(demands[i][:t + 1]) for t in range(horizon)] for i in range(nItems)]

# x[t] is the item produced at time t
x = VarArray(size=horizon, dom=range(nItems))

# p[i][t] is 1 if the ith item is produced at time t
p = VarArray(size=[nItems, horizon], dom={0, 1})

# z[t] is the changeover cost incurred at time t
z = VarArray(size=horizon - 1, dom=changeCosts)

satisfy(
    # channeling variables
    [p[i][t] == (x[t] == i) for i in range(nItems) for t in range(horizon)],

    # ensuring that deadlines of demands are respected
    [Sum(p[i][:t + 1]) >= required[i][t] for i in range(nItems) for t in range(horizon) if t == 0 or required[i][t - 1] != required[i][t]],

    # computing changeover costs
    [z[t] == changeCosts[x[t], x[t + 1]] for t in range(horizon - 1)],

    # tag(redundant-constraints)
    [Count(x, value=i) >= required[i][-1] for i in range(nItems)]
)

minimize(
    Sum(stockingCosts[i] * (Sum(p[i][:t + 1]) - required[i][t]) for i in range(nItems) for t in range(horizon))
    + Sum(z)
)

"""
1) we post a subset of constraints about deadlines only when a skip
2) should we compute Sum(p[i][:t + 1]) - required[i][t] with variables?
"""

