"""
See Vianney Coppé, Xavier Gillard, Pierre Schaus:
Solving the Constrained Single-Row Facility Layout Problem with Decision Diagrams. CP 2022: 14:1-14:18
"""

from pycsp3 import *

lengths, traffics = data
n = len(lengths)

ordered_lengths = sorted(lengths)

pairs = [(i, j) for i, j in combinations(n, 2) if i + 1 < j]

# x[i] is the department at the ith position
x = VarArray(size=n, dom=range(n))

# y[i] is the length of the department at the ith position
y = VarArray(size=n, dom=lengths)

# d[i][j] is the end-to-start distance between the ith and the jth departments
d = VarArray(size=[n, n], dom=lambda i, j: range(sum(ordered_lengths[:j - i - 1]), sum(ordered_lengths[i - j + 1:]) + 1) if i + 1 < j else None)

satisfy(
    # ensuring a linear ordering of the departments
    AllDifferent(x),

    # computing lengths
    [(x[i], y[i]) in [(j, lengths[j]) for j in range(n)] for i in range(n)],

    # computing distances
    [d[i][j] == Sum(y[i + 1:j]) for i, j in pairs]
)

if not variant():
    minimize(
        # minimizing weighted end-to-start distances
        Sum(d[i][j] * traffics[x[i], x[j]] for i, j in pairs)
    )

elif variant("mini"):
    z = VarArray(size=[n, n], dom=lambda i, j: traffics if i + 1 < j else None)

    m, M = min(v for row in traffics for v in row), max(v for row in traffics for v in row)

    w = VarArray(size=[n, n], dom=lambda i, j: range(sum(ordered_lengths[:j - i - 1]) * m, sum(ordered_lengths[i - j + 1:]) * M + 1) if i + 1 < j else None)

    satisfy(
        [(z[i][j], x[i], x[j]) in [(traffics[v1][v2], v1, v2) for v1 in range(n) for v2 in range(n)] for i, j in pairs],

        [w[i][j] == d[i][j] * z[i][j] for i, j in pairs]
    )

    minimize(
        # minimizing weighted end-to-start distances
        Sum(w)
    )

