"""
See Xavier Gillard, Pierre Schaus:
Large Neighborhood Search with Decision Diagrams. IJCAI 2022: 4754-4760
"""

from pycsp3 import *

distances, windows = data
Earliest, Latest = cp_array(zip(*windows))
horizon = max(Latest) + 1
n = len(distances)

# x[i] is the customer (node) visited in the ith position
x = VarArray(size=n + 1, dom=range(n))

# a[i] is the time when is visited the customer in the ith position
a = VarArray(size=n, dom=range(horizon))

satisfy(
    #  making it a tour while starting and ending at city 0
    [x[0] == 0, x[-1] == 0, a[0] == 0],

    AllDifferent(x[:-1]),

    # enforcing time windows
    [
        [Earliest[x[i]] <= a[x[i]] for i in range(n)],
        [a[x[i]] <= Latest[x[i]] for i in range(n)],
        [a[x[i + 1]] >= a[x[i]] + distances[x[i], x[i + 1]] for i in range(n - 1)]
    ]
)

minimize(
    # minimizing travelled distance
    Sum(distances[x[i], x[(i + 1) % n]] for i in range(n))
)
