"""
See Xavier Gillard, Pierre Schaus:
Large Neighborhood Search with Decision Diagrams. IJCAI 2022: 4754-4760
"""

from pycsp3 import *

distances, windows = data
horizon = max(latest for (_, latest) in windows) + 1
n = len(distances)

# x[i] is the node succeeding to the ith node
x = VarArray(size=n, dom=range(n))

# a[i] is the time when is visited the ith node
a = VarArray(size=n, dom=lambda i: range(windows[i][0], windows[i][1] + 1))

satisfy(
    #  making it a tour while starting and ending at city 0
    a[0] == 0,

    # avoiding self-loops
    [x[i] != i for i in range(n)],

    # forming a circuit
    Circuit(x),

    # enforcing time windows
    [(x[i] == 0) | (a[x[i]] >= a[i] + distances[i][x[i]]) for i in range(n)]
)

minimize(
    # minimizing travelled distance
    Sum(distances[i, x[i]] for i in range(n))
)
