"""
See "An Exercise for the Mind: A 10 by 10 Math Puzzle:
      A Pattern Recognition Game: Meditation on an Open Maze"
See https://chycho.blogspot.com/2014/01/an-exercise-for-mind-10-by-10-math.html

Example of Execution:
  python CalvinPuzzle.py -data=5
  python CalvinPuzzle.py -variant=table -data=5
"""

from pycsp3 import *

n = data or 5

# x[i][j] is the value in the grid at row i and column j
x = VarArray(size=[n, n], dom=range(1, n * n + 1))

# possible neighbours
offsets = [(-3, 0), (3, 0), (0, -3), (0, 3), (-2, -2), (-2, 2), (2, -2), (2, 2)]
N = [[[x[i + oi][j + oj] for (oi, oj) in offsets if 0 <= i + oi < n and 0 <= j + oj < n] for j in range(n)] for i in range(n)]

satisfy(
    # putting all values from 1 to n*n in the grid
    AllDifferent(x),

    # tag(symmetry-breaking)
    x[0][0] == 1
)

if not variant():
    satisfy(
        # each cell must be linked to its neighbors
        (x[i][j] == n * n) | Exist(y == x[i][j] + 1 for y in N[i][j]) for i in range(n) for j in range(n)
    )


elif variant("table"):

    def table(i, j):
        r = len(N[i][j]) + 1
        return [tuple(k if i == 0 else (k + 1) if i == j else ANY for i in range(r)) for k in range(1, n * n) for j in range(1, r)] \
            + [(n * n, *[ANY] * (r - 1))]


    satisfy(
        # each cell must be linked to its neighbors
        (x[i][j], N[i][j]) in table(i, j) for i in range(n) for j in range(n)
    )

"""
1) using an hybrid table is possible
"""
