"""
See https://www.cs.york.ac.uk/aig/projects/implied/docs/CPAIOR03.pdf
English style (standard), 33 holes

python PegSolitaire.py -data=[3,3,0] -variant=english
"""

from pycsp3 import *

from PegSolitaire_Generator import generate_boards, build_transitions

assert variant() in {"english", "french"}

origin_x, origin_y, nMoves = data

init_board, final_board = generate_boards(variant(), origin_x, origin_y)
n, m = len(init_board), len(init_board[0])
transitions = build_transitions(init_board)
nTransitions = len(transitions)

horizon = sum(sum(v for v in row if v) for row in init_board) - sum(sum(v for v in row if v) for row in final_board)
nMoves = horizon if nMoves <= 0 or horizon < nMoves else nMoves
assert 0 < nMoves <= horizon

pairs = [(i, j) for i in range(n) for j in range(m) if init_board[i][j] is not None]

# x[i,j,t] is the value at row i and column j at time t
x = VarArray(size=[nMoves + 1, n, m], dom=lambda t, i, j: {0, 1} if init_board[i][j] is not None else None)

# y[t] is the move (transition) performed at time t
y = VarArray(size=nMoves, dom=range(nTransitions))

satisfy(
    # setting the initial board
    x[0] == init_board,

    # setting the final board
    x[-1] == final_board
)


def unchanged(i, j, t):
   valid = [k for k, tr in enumerate(transitions) if (i, j) in (tr[0:2], tr[2:4], tr[4:6])]
   return None if len(valid) == 0 else iff(conjunction(y[t] != k for k in valid), x[t, i, j] == x[t + 1, i, j])


def to0(i, j, t):
   valid = [k for k, tr in enumerate(transitions) if (i, j) in (tr[0:2], tr[2:4])]
   return None if len(valid) == 0 else iff(disjunction(y[t] == k for k in valid), (x[t, i, j] == 1) & (x[t + 1, i, j] == 0))


def to1(i, j, t):
   valid = [k for k, tr in enumerate(transitions) if (i, j) == tr[4:6]]
   return None if len(valid) == 0 else iff(disjunction(y[t] == k for k in valid), (x[t, i, j] == 0) & (x[t + 1, i, j] == 1))

satisfy(
   [unchanged(i, j, t) for (i, j) in pairs for t in range(nMoves)],
   [to0(i, j, t) for (i, j) in pairs for t in range(nMoves)],
   [to1(i, j, t) for (i, j) in pairs for t in range(nMoves)]
)


""" Comments
1) x[0] == init_board is possible because cells with None (either in variables or values) are discarded
   This is equivalent to [x[0][i][j] == init_board[i][j] for (i, j) in pairs]
"""
