"""
LPCP contest 2022 (Problem 5)
See https://github.com/lpcp-contest/lpcp-contest-2022/tree/main/problem-5

Example of Execution:
  python Slant.py -data=instance01.in -dataparser=Slant_Parser.py
"""

from pycsp3 import *

grid = data
n = len(grid)

DOWN_DIAG, UP_DIAG = 0, 1

# e[k][l] is 1 if the edge in the intermediate cell at coordinates (k,l) is an upward downward, 0 if a downward diagonal
e = VarArray(size=[n - 1, n - 1], dom={DOWN_DIAG, UP_DIAG})

# x[i][j] is the number of connected nodes (effective neighbors, or bridges) to node (i,j)
x = VarArray(size=[n, n], dom=lambda i, j: {grid[i][j]} if grid[i][j] != -1 else range(5))

# d[i][j] is the distance of node (i,j) from the root to the chain/tree it belongs
d = VarArray(size=[n, n], dom=range(n * n + 1))


def connections_of(i, j):
    # returns a list of tuples (k,l,a,ii,jj) where (ii,jj) is a possible node that can be reached from (i,j)
    # by using a diagonal a on the edge (k,l)
    t = []
    if i > 0:
        if j > 0:
            t.append((i - 1, j - 1, DOWN_DIAG, i - 1, j - 1))
        if j < n - 1:
            t.append((i - 1, j, UP_DIAG, i - 1, j + 1))
    if i < n - 1:
        if j > 0:
            t.append((i, j - 1, UP_DIAG, i + 1, j - 1))
        if j < n - 1:
            t.append((i, j, DOWN_DIAG, i + 1, j + 1))
    return t


nodes = [(i, j) for i in range(n) for j in range(n)]
connections = [connections_of(i, j) for i, j in nodes]

satisfy(
    # computing the number of neighbors
    [x[i][j] == Sum(e[k][l] == a for (k, l, a, _, _) in connections_of(i, j)) for i, j in nodes],

    # isolated nodes are roots (are a distance 0)
    [(x[i][j] != 0) | (d[i][j] == 0) for i, j in nodes],

    # nodes with at least two neighbors cannot be roots (be at distance 0)
    [(x[i][j] <= 1) | (d[i][j] != 0) for i, j in nodes],

    # nodes with at least two neighbors have exactly one parent (a node with the same distance minus 1)
    [(x[i][j] <= 1) | (Count((e[k][l] == a) & (d[i][j] == d[ii][jj] + 1) for (k, l, a, ii, jj) in connections_of(i, j)) == 1) for i, j in nodes],

    # the distance between any two neighbors is always 1
    [(e[k][l] != a) | (abs(d[i][j] - d[ii][jj]) == 1) for i, j in nodes for (k, l, a, ii, jj) in connections_of(i, j)]
)

"""
A model with tables is possible (for replacing the 4 last posted groups of constraints). To be tested
"""
